import greenfoot.*;
import java.awt.Color;
import java.util.List;

public class Object
{
    public vector pos = new vector(0f,0f,0f);
    public vector rot = new vector(0f,0f,0f);
    public vector[] verts;
    public vector[] saveVerts;
    public int[][] faces;
    public vector[] normals;
    public Color[] colors;
    public Object(vector vPos, vector vRot, vector[] vVerts, int[][] vFaces, vector[] vNormals, Color[] cColors)//Faces[0 through 2]= vertices, Faces[3]= face normal
    {
        pos = vPos;
        rot = vRot;
        verts = vVerts;
        saveVerts = vVerts;
        normals = vNormals;
        faces = new int[vFaces.length][4];
        if(vFaces[0].length == 6)
        {
            for(int i = 0;i<vFaces.length;i++)
                faces[i] = new int[]{vFaces[i][0],vFaces[i][2],vFaces[i][4],vFaces[i][5]};
        }
        else
            faces = vFaces;
        
        colors = new Color[faces.length];
        if(cColors.length == 1)
        {
            for(int i = 0;i<vFaces.length;i++)
                colors[i] = cColors[0];
        }
        else
            colors = cColors;
    }
    
    private vector centroid(vector[] face)
    {
        return new vector((face[0].x+face[1].x+face[2].x)/3,(face[0].y+face[1].y+face[2].y)/3,(face[0].z+face[1].z+face[2].z)/3);
    }
    
    private float faceDist(vector[] face)
    {
        return vector.Dist(Universe.view.pos,centroid(face));
    }
    
    private int[] raterizePoint(vector point)
    {
        float dpX = 0;
        float dpY = 0;
        if(point.z != 0)
        {
            dpX = (point.x-(Universe.view.pos.x-point.z))/(2*point.z);
            dpY = (point.y-(Universe.view.pos.y-point.z))/(2*point.z);
        }
        return new int[]{Math.round(dpX*512),Math.round(dpY*512)};
    }
    
    public void render(GreenfootImage display)
    {
        vector[][] vFaces = new vector[faces.length][4];
            
        //Populate faces array
        for(int i = 0;i<faces.length;i++)
        {
            //System.out.println("i="+i+"\t"+(faces[i][0]-1)+"\t"+(faces[i][1]-1)+"\t"+(faces[i][2]-1));
            //System.out.println(""+(faces[i][0]-1));
            vFaces[i] = new vector[]{vector.Add(verts[faces[i][0]-1],pos),
                                    vector.Add(verts[faces[i][1]-1],pos),
                                    vector.Add(verts[faces[i][2]-1],pos),
                                    normals[faces[i][3]-1]};
        }
        //System.out.println(""+rot.x+","+rot.y+","+rot.z);
        Color[] colors = this.colors;
        
        float faceDists[][] = new float[vFaces.length][2];
        for(int i = 0;i<vFaces.length;i++)
        {
            faceDists[i] = new float[]{faceDist(vFaces[i]),i};
        }
        for(int i=faceDists.length;i>1;i--)
            fnSortHeap(faceDists,i - 1);
            
        vector[][] tempFaces = new vector[vFaces.length][4];
        Color[] tempColors = new Color[vFaces.length];
        
        //Transfer sortedness to array
        for(int i = 0;i<faces.length;i++)
        {
            tempFaces[i] = vFaces[(int)faceDists[i][1]];
            tempColors[i] = colors[(int)faceDists[i][1]];
            //System.out.println((int)faceDists[i][1]+"\t"+faceDists[i][0]);
        }
        
//             for(int i = 0;i<faces.length;i++)
//                 System.out.println(""+tempFaces[i][0].x+"\t"+faces[i][0].x+"\t"+colors[i]);
        
        vFaces = tempFaces;
        colors = tempColors;
        
        //Render face
        for(int i = vFaces.length-1;i>=0;i--)
        {
            float normZ = vFaces[i][3].z;
            if(normZ < 0.25) //Backface culling
            {
                if(normZ > -0.25)
                    normZ = -0.25f;
                display.setColor(new Color((int)(colors[i].getRed()*(-normZ)),(int)(colors[i].getGreen()*(-normZ)),(int)(colors[i].getBlue()*(-normZ))));
                display.fillPolygon(new int[]{raterizePoint(vFaces[i][0])[0],raterizePoint(vFaces[i][1])[0],raterizePoint(vFaces[i][2])[0]},
                                    new int[]{raterizePoint(vFaces[i][0])[1],raterizePoint(vFaces[i][1])[1],raterizePoint(vFaces[i][2])[1]},3);
            }
        }
    }
    
    public static float[][] fnSortHeap(float array[][], int arr_ubound) //http://www.roseindia.net/java/beginners/arrayexamples/heapSort.shtml
    {
        int i, o;
        int lChild, rChild, mChild, root;
        float[] temp;
        root = (arr_ubound-1)/2;
        
        for(o = root; o >= 0; o--)
        {
            for(i=root;i>=0;i--)
            {
                lChild = (2*i)+1;
                rChild = (2*i)+2;
                if((lChild <= arr_ubound) && (rChild <= arr_ubound))
                {
                    if(array[rChild][0] >= array[lChild][0])
                        mChild = rChild;
                    else
                        mChild = lChild;
                }
                else
                {
                    if(rChild > arr_ubound)
                        mChild = lChild;
                    else
                        mChild = rChild;
                }
                
                if(array[i][0] < array[mChild][0])
                {
                    temp = array[i];
                    array[i] = array[mChild];
                    array[mChild] = temp;
                }
            }
        }
        temp = array[0];
        array[0] = array[arr_ubound];
        array[arr_ubound] = temp;
        return array;
    }
}
